Canvas = class(Group);

function Canvas:Constructor()
    Group.Constructor(self, "Turbine.UI.Window");
    self.idStr = "new Canvas";

    self.selectedShapes = {};

    self.editor = Editor(self);
    local button = Turbine.UI.Control();
    button:SetSize(16, 16);
    button:SetParent(self);
    button:SetZOrder(2147483647);
    button.UpdateBackground = function(b)
        if (b:IsMouseInside()) then
            b:SetBackground(0x41000198);
        else
            b:SetBackground(0x41000196);
        end
    end
    button.MouseEnter = function(b)
        b:UpdateBackground();
    end
    button.MouseLeave = function(b)
        b:UpdateBackground();
    end
    button.IsMouseInside = function(b)
        local x, y = b:GetMousePosition();
        return (x >= 0) and (x < 16) and (y >= 0) and (y < 16);
    end
    button.MouseClick = function(b)
        DoCallbacks(self, "CloseButtonClicked", args);
        self:SetVisible(false);
    end
    Thurallor.UI.Tooltip("Minimize window"):Attach(button);
    self.closeButton = button;

    -- Hardcoded properties
    self.properties.Class = Canvas;
    self.changedProperties.Class = Canvas;
    self.resizable = false; -- resizable, but not with a GrabPoint
    self.baseClass.SetVisible(self, true);
    self.baseClass.SetZOrder(self, -1);

    -- Default properties
    self.applyingRemoteChanges = true;
    self:SetSize(620, 360);
    self:SetProperty("Opacity", 0.5);
    self:SetBackColor(Turbine.UI.Color.Black);
    self:SetLocked(true);
    self.applyingRemoteChanges = false;
    
    self:SavePropertyDefaults();

    self.sizer = Thurallor.UI.Resizer(self);
end

function Canvas:MouseMove()
    Shape.MouseMove(self);
    -- Later: Add multi-selecting by drawing a rectangle around shapes.
end

function Canvas:MouseEnter()
    Shape.MouseEnter(self);
    self.closeButton:SetVisible(true);
    self.closeButton:UpdateBackground();
end

function Canvas:MouseLeave()
    Shape.MouseLeave(self);
    self.closeButton:UpdateBackground();
    if (not self.closeButton:IsMouseInside()) then
        self.closeButton:SetVisible(false);
    end
end

function Canvas:MouseClick(args)
    -- When the user clicks on the background behind the shapes
    self:Deselect();
    self:Focus();
end

function Canvas:AddShapeAt(shape, left, top)
    Group.AddShapeAt(self, shape, left, top);
    shape.prevLeft, shape.prevTop = left or 0, top or 0;
    AddCallback(shape, "SizeChanged", function(s)
        self:ShapeSizeChanged(s);
    end);
    AddCallback(shape, "PositionChanged", function(s)
        self:ShapePositionChanged(s);
        s.prevLeft, s.prevTop = s:GetPosition();
    end);
    AddCallback(shape, "MouseClick", function(s, args)
        if (s.mouseMovedAfterMouseDown) then
            -- Not a real "click"; we shouldn't have gotten this event.
            return;
        end
        if (not self:IsShiftKeyDown()) then
            self:Deselect();
        end
-- to do: check for ctrl key?
        self:SelectShape(s);
    end);
end

function Canvas:RemoveShape(shape)
    if (self.selectedShapes[shape]) then
        self:Deselect(shape);
    end
    Group.RemoveShape(self, shape);
end

function Canvas:CloneShape(shape)
    Group.CloneShape(self, shape);
    self:Reselect();
end

function Canvas:ApplyChanges(properties, from)
    Shape.ApplyChanges(self, properties, from);

    -- Selection may have changed.
    self:Reselect();
    
    -- If locked, rob focus from TextShapes
--    if (self:IsLocked()) then
--        self:Focus();
--    end
end

function Canvas:Reselect()
    -- Scan all shapes for IsSelected(), reapply their selection, and reconstruct the self.selectedShapes table.
    self.selectedShapes = {};
    for _, shape in pairs(self.shapesByIDStr) do
        if (shape:IsSelected()) then
            shape:SetSelected(true);
            self.selectedShapes[shape] = true;
        else
            shape:SetSelected(false);
            self.selectedShapes[shape] = nil;
        end
    end
    DoCallbacks(self, "SelectionChanged");
end

function Canvas:Deselect(shape)
    if (not shape) then
        -- deselect all selected shapes
        for shape, _ in pairs(self.selectedShapes) do
            self:Deselect(shape);
        end
    else
        shape:SetSelected(false);
        self.selectedShapes[shape] = nil;
    end
    DoCallbacks(self, "SelectionChanged");
end

-- Adds a shape to the current selection and puts a marquee on it
function Canvas:SelectShape(shape)
    self.selectedShapes[shape] = true;
    shape:SetSelected(true);
    DoCallbacks(self, "SelectionChanged");
end

function Canvas:GetSelection()
    return self.selectedShapes;
end

function Canvas:ShapeIsSelected(shape)
    return shape:IsSelected();
end

function Canvas:ShowContextMenu()
    -- Currently there's no context menu when you right-click on a canvas.
end

function Canvas:ShowSelectionContextMenu()
    local firstShape = next(self.selectedShapes);
    if (firstShape) then
        local secondShape = next(self.selectedShapes, firstShape);
        if (secondShape) then
            -- Multiple shapes selected
            -- to do: show things like align, distribute, etc. that apply to multiple shapes
        else
            -- Single shape selected
            firstShape:GetContextMenu():ShowMenu();
        end
    end
end

function Canvas:SetShape(str, from)
    Group.SetShape(self, str, from);
    self:Reselect();
end

function Canvas:ShapeSizeChanged(shape)
    if (self:ShapeIsSelected(shape)) then
        local width, height = shape:GetSize();
        shape.marquee:SetSize(width + 2, height + 2);
    end
end

function Canvas:ShapePositionChanged(shape)
    if (self:ShapeIsSelected(shape)) then
        local left, top = shape:GetPosition();
        shape.marquee:SetPosition(left - 1, top - 1);
        
        -- If other shapes are also selected, move them by the same amount
        if (not shape.doingGroupMove) then
            local xShift = left - shape.prevLeft;
            local yShift = top - shape.prevTop
            for s, _ in pairs(self.selectedShapes) do
                if (s ~= shape) then
                    local x, y = s:GetPosition();
                    s.doingGroupMove = true;
                    s:SetPosition(x + xShift, y + yShift);
                end
            end
        end
        shape.doingGroupMove = false;
    end
end

function Canvas:ImportShapes(str)
    Group.ImportShapes(self, str);
    self:Reselect();
end

function Canvas:SetEditorEnabled(enabled)
    self.editor:SetVisible(enabled);
    return self.editor;
end

function Canvas:Activate()
    self.baseClass.Activate(self);
end

function Canvas:SetVersion(version) -- plugin version used to create this Canvas, times 100
    self:SetProperty("Version", math.floor(version * 100));
end

function Canvas:GetVersion()
    return self.properties.Version / 100;
end


-- Standard untracked properties
function Canvas:SetVisible(visible)
    self.baseClass.SetVisible(self, visible);
    self.editor:SetVisible(visible);
    if (visible) then
        self:SetLocked(self.properties.Locked);
    end
end

function Canvas:SetZOrder(zOrder)
    -- Override Shape.SetZOrder because we don't want to track it for Canvases
    self.baseClass.SetZOrder(self, zOrder);
end

function Canvas:GetZOrder()
    return self.baseClass.GetZOrder(self);
end

function Canvas:SetPosition(left, top)
    -- Override Shape.SetPosition because we don't want to track it for Canvases
    self.baseClass.SetPosition(self, left, top);
end

function Canvas:GetPosition()
    return self.baseClass.GetPosition(self);
end

function Canvas:SetLeft(left)
    -- Override Shape.SetLeft because we don't want to track it for Canvases
    self.baseClass.SetLeft(self, left);
end

function Canvas:GetLeft()
    return self.baseClass.GetLeft(self);
end

function Canvas:SetTop(top)
    -- Override Shape.SetTop because we don't want to track it for Canvases
    self.baseClass.SetTop(self, top);
end

function Canvas:GetTop()
    return self.baseClass.GetTop(self);
end


-- Standard tracked properties
function Canvas:_SizeChanged()
    local width, height = self:GetSize();
    self:SetProperty("Width", width);
    self:SetProperty("Height", height);
    if (self.closeButton) then
        self.closeButton:SetLeft(width - 16);
    end
    -- Perhaps later: Adjust child-clipping, position of children, scrollbar positions
end

function Canvas:SetText(text) -- not displayed; used only for filename
    self:SetProperty("Text", text);
end

function Canvas:GetText()
    return self.properties.Text;
end

function Canvas:SetBackColor(color)
    color.A = self.properties.Opacity;
    self.baseClass.SetBackColor(self, color);
    color.A = 1;
    self:SetProperty("BackColor", color);
end

function Canvas:GetBackColor()
    return self.properties.BackColor;
end

function Canvas:SetOpacity(opacity) -- Sets the background color opacity, rather than the opacity of the window
    local color = self:GetBackColor();
    color.A = opacity;
    self:SetBackColor(color);
    self:SetProperty("Opacity", opacity);
end

function Canvas:GetOpacity()
    return self.properties.Opacity;
end

-- Custom tracked properties
function Canvas:SetLocked(locked)
    self:SetProperty("Locked", locked);
    if (locked and not self:AmOwner()) then
        -- Disable editing
        self:SetEditorEnabled(false);
        self.editor:MaskCanvas(true);
    else
        -- Enable editing
        self:SetEditorEnabled(true);
        self.editor:MaskCanvas(false);
    end
end

function Canvas:IsLocked(locked)
    return self.properties.Locked;
end

function Canvas:SetOwner(owner)
    Shape.SetOwner(self, owner);
    self:SetLocked(self.properties.Locked);
end
